"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
const aws_1 = __importDefault(require("./aws"));
const utils_1 = __importDefault(require("./utils"));
const Error_1 = __importDefault(require("./Error"));
const ModelStore_1 = __importDefault(require("./ModelStore"));
var TransactionReturnOptions;
(function (TransactionReturnOptions) {
    TransactionReturnOptions["request"] = "request";
    TransactionReturnOptions["documents"] = "documents";
})(TransactionReturnOptions || (TransactionReturnOptions = {}));
var TransactionType;
(function (TransactionType) {
    TransactionType["get"] = "get";
    TransactionType["write"] = "write";
})(TransactionType || (TransactionType = {}));
module.exports = (transactions, settings = { "return": TransactionReturnOptions.documents }, callback) => {
    if (typeof settings === "function") {
        callback = settings;
        settings = { "return": TransactionReturnOptions.documents };
    }
    if (typeof transactions === "function") {
        callback = transactions;
        transactions = null;
    }
    const promise = (() => __awaiter(void 0, void 0, void 0, function* () {
        if (!Array.isArray(transactions) || transactions.length <= 0) {
            throw Error_1.default.InvalidParameter("You must pass in an array with items for the transactions parameter.");
        }
        const transactionObjects = yield Promise.all(transactions);
        const transactionParams = {
            "TransactItems": transactionObjects
        };
        if (settings.return === TransactionReturnOptions.request) {
            return transactionParams;
        }
        let transactionType;
        if (settings.type) {
            switch (settings.type) {
                case TransactionType.get:
                    transactionType = "transactGetItems";
                    break;
                case TransactionType.write:
                    transactionType = "transactWriteItems";
                    break;
                default:
                    throw Error_1.default.InvalidParameter("Invalid type option, please pass in \"get\" or \"write\".");
            }
        }
        else {
            transactionType = transactionObjects.map((a) => Object.keys(a)[0]).every((key) => key === "Get") ? "transactGetItems" : "transactWriteItems";
        }
        const modelNames = transactionObjects.map((a) => Object.values(a)[0].TableName);
        const uniqueModelNames = utils_1.default.unique_array_elements(modelNames);
        const models = uniqueModelNames.map((name) => ModelStore_1.default(name));
        models.forEach((model, index) => {
            if (!model) {
                throw Error_1.default.InvalidParameter(`Model "${uniqueModelNames[index]}" not found. Please register the model with dynamoose before using it in transactions.`);
            }
        });
        yield Promise.all(models.map((model) => model.pendingTaskPromise()));
        // TODO: remove `as any` here (https://stackoverflow.com/q/61111476/894067)
        const result = yield aws_1.default.ddb()[transactionType](transactionParams).promise();
        return result.Responses ? yield Promise.all(result.Responses.map((item, index) => {
            const modelName = modelNames[index];
            const model = models.find((model) => model.name === modelName);
            return (new model.Document(item.Item, { "fromDynamo": true })).conformToSchema({ "customTypesDynamo": true, "checkExpiredItem": true, "saveUnknown": true, "type": "fromDynamo" });
        })) : null;
    }))();
    if (callback) {
        promise.then((result) => callback(null, result)).catch((error) => callback(error));
    }
    else {
        return promise;
    }
};
